"""
Debug eager mode
"""


_debug_eager_mode_enabled = False


def enable_debug_eager_mode():
    """
    For debugging.

    Enables TF eager mode.
    Also, all layers will directly be created, and then due to TF eager mode directly evaluated.
    """
    global _debug_eager_mode_enabled
    import tensorflow as tf

    tf.compat.v1.enable_eager_execution()
    _debug_eager_mode_enabled = True


def disable_debug_eager_mode():
    """
    For debugging.

    Enables TF eager mode.
    Also, all layers will directly be created, and then due to TF eager mode directly evaluated.
    """
    global _debug_eager_mode_enabled
    import tensorflow as tf

    tf.compat.v1.disable_eager_execution()
    _debug_eager_mode_enabled = False


def is_debug_eager_mode_enabled() -> bool:
    """
    :return: True if debug eager mode is enabled.
    """
    return _debug_eager_mode_enabled



"""
Some mathematical functions, in pure NumPy.
"""

from __future__ import annotations
from typing import Union, Dict
import numpy


def next_power_of_two(n: int) -> int:
    """next power of two, >= n"""
    return 2 ** (int(n - 1).bit_length())


class PiecewiseLinear:
    """
    Piecewise linear function.
    """

    def __init__(self, values: Dict[Union[int, float], Union[int, float]]):
        self._sorted_items = sorted(values.items())
        self._sorted_keys = numpy.array([x for x, _ in self._sorted_items])
        self._sorted_values = numpy.array([y for _, y in self._sorted_items])

    def __call__(self, x: Union[int, float]) -> Union[int, float]:
        steps = self._sorted_keys
        values = self._sorted_values
        return numpy.interp(x, steps, values)



"""
Provide basic Byte-Pair-Encoding (BPE) utilities.
"""

from __future__ import annotations
from typing import Optional, List, Dict, Callable
import re
import numpy


BpePostMergeSymbol = "@@"


class StandardBytePairEncoder:
    """
    Code is partly taken from subword-nmt/apply_bpe.py.
    Author: Rico Sennrich, code under MIT license.

    Use operations learned with learn_bpe.py to encode a new text.
    The text will not be smaller, but use only a fixed vocabulary, with rare words
    encoded as variable-length sequences of subword units.

    Reference:
    Rico Sennrich, Barry Haddow and Alexandra Birch (2016). Neural Machine Translation of Rare Words with Subword Units.
    Proceedings of the 54th Annual Meeting of the Association for Computational Linguistics (ACL 2016). Berlin, Germany.

    """

    def __init__(self, bpe_codes_file, labels=None):
        """
        :param str bpe_codes_file: codes file
        :param list[str]|None labels: vocab
        """
        self.labels = labels
        self._load(bpe_codes_file)
        self._bpe_encode_cache = {}
        self._bpe_separator = BpePostMergeSymbol

    _file_cache = {}  # filename -> bpe_file_version, bpe_codes, bpe_codes_reverse

    def _load(self, bpe_codes_file):
        """
        Load BPE codes from file

        :param str bpe_codes_file: codes file
        """
        if bpe_codes_file in self._file_cache:
            self._bpe_file_version, self._bpe_codes, self._bpe_codes_reverse = self._file_cache[bpe_codes_file]
            return
        # check version information
        bpe_file_first_line = open(bpe_codes_file, "r").readline()
        if bpe_file_first_line.startswith("#version:"):
            self._bpe_file_version = tuple(
                [int(x) for x in re.sub(r"(\.0+)*$", "", bpe_file_first_line.split()[-1]).split(".")]
            )
        else:
            self._bpe_file_version = (0, 1)
        self._bpe_codes = [
            tuple(item.split()) for item in open(bpe_codes_file, "rb").read().decode("utf8").splitlines()
        ]
        # some hacking to deal with duplicates (only consider first instance)
        self._bpe_codes = dict([(code, i) for (i, code) in reversed(list(enumerate(self._bpe_codes)))])
        self._bpe_codes_reverse = dict([(pair[0] + pair[1], pair) for pair, i in self._bpe_codes.items()])
        self._file_cache[bpe_codes_file] = (self._bpe_file_version, self._bpe_codes, self._bpe_codes_reverse)

    @staticmethod
    def _get_pairs(word):
        """
        :param tuple[str] word: represented as tuple of symbols (symbols being variable-length strings)
        :return: set of symbol pairs in a word
        :rtype: set[(str,str)]
        """
        pairs = set()
        prev_char = word[0]
        for char in word[1:]:
            pairs.add((prev_char, char))
            prev_char = char
        return pairs

    def _encode_word(self, orig):
        """
        Encode word based on list of BPE merge operations, which are applied consecutively.
        :param str orig:
        :rtype: tuple[str]
        """

        if orig in self._bpe_encode_cache:
            return self._bpe_encode_cache[orig]

        if self._bpe_file_version == (0, 1):
            word = tuple(orig) + ("</w>",)
        elif self._bpe_file_version == (0, 2):  # more consistent handling of word-final segments
            word = tuple(orig[:-1]) + (orig[-1] + "</w>",)
        else:
            raise NotImplementedError

        pairs = self._get_pairs(word)
        if not pairs:
            return orig

        while True:
            bigram = min(pairs, key=lambda pair: self._bpe_codes.get(pair, float("inf")))
            if bigram not in self._bpe_codes:
                break
            first, second = bigram
            new_word = []
            i = 0
            while i < len(word):
                try:
                    j = word.index(first, i)
                    new_word.extend(word[i:j])
                    i = j
                except ValueError:
                    new_word.extend(word[i:])
                    break

                if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                    new_word.append(first + second)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            new_word = tuple(new_word)
            word = new_word
            if len(word) == 1:
                break
            else:
                pairs = self._get_pairs(word)

        # don't print end-of-word symbols
        if word[-1] == "</w>":
            word = word[:-1]
        elif word[-1].endswith("</w>"):
            word = word[:-1] + (word[-1].replace("</w>", ""),)

        if self.labels:
            word = self._check_vocab_and_split(word, self._bpe_codes_reverse, self.labels, self._bpe_separator)

        self._bpe_encode_cache[orig] = word
        return word

    def _check_vocab_and_split(self, orig, bpe_codes, vocab, separator):
        """
        Check for each segment in word if it is in-vocabulary,
        and segment OOV segments into smaller units by reversing the BPE merge operations
        """

        out = []

        for segment in orig[:-1]:
            if segment + separator in vocab:
                out.append(segment)
            else:
                # sys.stderr.write('OOV: {0}\n'.format(segment))
                for item in self._recursive_split(segment, bpe_codes, vocab, separator, False):
                    out.append(item)

        segment = orig[-1]
        if segment in vocab:
            out.append(segment)
        else:
            # sys.stderr.write('OOV: {0}\n'.format(segment))
            for item in self._recursive_split(segment, bpe_codes, vocab, separator, True):
                out.append(item)

        return out

    def _recursive_split(self, segment, bpe_codes, vocab, separator, final=False):
        """Recursively split segment into smaller units (by reversing BPE merges)
        until all units are either in-vocabulary, or cannot be split further."""

        # noinspection PyBroadException
        try:
            if final:
                left, right = bpe_codes[segment + "</w>"]
                right = right[:-4]
            else:
                left, right = bpe_codes[segment]
        except Exception:  # TODO fix
            # sys.stderr.write('cannot split {0} further.\n'.format(segment))
            yield segment
            return

        if left + separator in vocab:
            yield left
        else:
            for item in self._recursive_split(left, bpe_codes, vocab, separator, False):
                yield item

        if (final and right in vocab) or (not final and right + separator in vocab):
            yield right
        else:
            for item in self._recursive_split(right, bpe_codes, vocab, separator, final):
                yield item

    def segment_sentence(self, sentence):
        """
        Segment single sentence (whitespace-tokenized string) with BPE encoding.

        :param str sentence:
        :rtype: list[str]
        """

        output = []

        found_category = False
        skip_category = False

        for word in sentence.split():
            if word[0] == "$" and len(word) > 1:
                found_category = True
                output.append(word)
            elif found_category is True and word[0] == "{":
                skip_category = True
                output.append(word)
            elif skip_category is True and word[0] != "}":
                output.append(word)
            else:
                found_category = False
                skip_category = False
                new_word = self._encode_word(word)

                for item in new_word[:-1]:
                    output.append(item + self._bpe_separator)
                output.append(new_word[-1])

        return output


class PrefixTree:
    """
    Prefix tree / trie.
    This class represents both a single node and the tree.
    """

    def __init__(self, prefix: str = "", root: Optional[PrefixTree] = None):
        """
        :param prefix:
        :param root:
        """
        self.prefix = prefix
        self.arcs: Dict[str, PrefixTree] = {}  # single char (or BpePostMergeSymbol) -> sub tree
        self.finished = False  # word finished here
        self.bpe_finished = False  # partial word finished here with BpePostMergeSymbol at end
        self.is_root = not root
        self.root = root

    def add(self, postfix: str, root: Optional[PrefixTree] = None) -> PrefixTree:
        """
        :param postfix:
        :param root:
        """
        if not root:
            if self.is_root:
                root = self
            else:
                assert self.root
                root = self.root
        self_ = self
        while True:
            if postfix == BpePostMergeSymbol:
                arc = postfix
                postfix_ = ""
            else:
                arc = postfix[:1]
                postfix_ = postfix[1:]
            if arc in self_.arcs:
                child = self_.arcs[arc]
            else:
                child = PrefixTree(root=root, prefix=self_.prefix + arc)
                self_.arcs[arc] = child
            if arc == BpePostMergeSymbol and not postfix_:
                self_.bpe_finished = True
            if postfix_:
                self_ = child
                postfix = postfix_
            else:
                child.finished = True
                return child


class Hyp:
    """
    Represents a hypothesis in the search.
    """

    def __init__(self, bpe_sym_history, cur_node):
        """
        :param list[str] bpe_sym_history:
        :param PrefixTree cur_node:
        """
        self.bpe_sym_history = bpe_sym_history
        self.cur_node = cur_node


class CharSyncSearch:
    """
    Covers the search hyps and the search itself.
    """

    def __init__(self, bpe: PrefixTree, word: str, word_pos: int = 0):
        """
        :param bpe:
        :param word:
        :param word_pos:
        """
        self.bpe = bpe
        self.word = word
        self.word_pos = word_pos
        self.hyps: List[Hyp] = [Hyp(bpe_sym_history=[], cur_node=bpe)]
        self.final_bpe_seqs: Optional[List[List[str]]] = None

    def _get_finished(self):
        assert self.word_pos == len(self.word)
        finals: List[List[str]] = []
        for hyp in self.hyps:
            if hyp.cur_node.finished:
                finals.append(hyp.bpe_sym_history + [hyp.cur_node.prefix])
        self.final_bpe_seqs = finals

    def _expand(self):
        assert self.word_pos < len(self.word)
        char = self.word[self.word_pos]
        new_hyps: List[Hyp] = []
        for hyp in self.hyps:
            if hyp.cur_node.bpe_finished:
                # Start again from root node.
                next_node = self.bpe.arcs.get(char)
                if next_node:
                    new_hyps.append(
                        Hyp(
                            bpe_sym_history=hyp.bpe_sym_history + [hyp.cur_node.prefix + BpePostMergeSymbol],
                            cur_node=next_node,
                        )
                    )
            next_node = hyp.cur_node.arcs.get(char)
            if next_node:
                new_hyps.append(Hyp(bpe_sym_history=hyp.bpe_sym_history, cur_node=next_node))
        self.hyps = new_hyps

    def search(self) -> List[List[str]]:
        """
        :return: collection of possible BPE symbol seqs
        """
        while self.word_pos < len(self.word):
            self._expand()
            self.word_pos += 1
        self._get_finished()
        return self.final_bpe_seqs


class HypInPos:
    """
    Represents a hypothesis in the search.
    """

    def __init__(self, bpe_sym_history, cur_node, pos):
        """
        :param list[str] bpe_sym_history:
        :param PrefixTree cur_node:
        :param int pos:
        """
        self.bpe_sym_history = bpe_sym_history
        self.cur_node = cur_node
        self.pos = pos


class DepthFirstSearch:
    """
    Depth-first search.
    """

    def __init__(self, bpe: PrefixTree, word: str, sampler: Optional[Callable[[], bool]] = None):
        """
        :param bpe:
        :param word:
        :param sampler:
        """
        self.bpe = bpe
        self.word = word
        self.sampler = sampler
        self.hyps: List[HypInPos] = []
        self.final_bpe_seq: Optional[List[str]] = None
        self._add_hyp(HypInPos(bpe_sym_history=[], cur_node=bpe, pos=0))

    def _add_hyp(self, hyp: HypInPos):
        if hyp.pos >= len(self.word):
            if hyp.cur_node.finished:
                self.final_bpe_seq = hyp.bpe_sym_history + [hyp.cur_node.prefix]
        else:
            self.hyps.append(hyp)

    def _expand(self):
        hyp = self.hyps.pop(-1)

        # Now check for possible further hyps.
        char = self.word[hyp.pos]
        new_hyps: List[HypInPos] = []
        if hyp.cur_node.bpe_finished:
            # Start again from root node.
            next_node = self.bpe.arcs.get(char)
            if next_node:
                new_hyps.append(
                    HypInPos(
                        bpe_sym_history=hyp.bpe_sym_history + [hyp.cur_node.prefix + BpePostMergeSymbol],
                        cur_node=next_node,
                        pos=hyp.pos + 1,
                    )
                )
        next_node = hyp.cur_node.arcs.get(char)
        if next_node:
            new_hyps.append(HypInPos(bpe_sym_history=hyp.bpe_sym_history, cur_node=next_node, pos=hyp.pos + 1))

        # Note that the order we check them will make this a depth-first or breadth-first search.
        if self.sampler and self.sampler():
            new_hyps = list(reversed(new_hyps))
        for hyp in new_hyps:
            self._add_hyp(hyp)

    def search(self) -> Optional[List[str]]:
        """
        :return: BPE symbol seq if one is found
        """
        while self.hyps and self.final_bpe_seq is None:
            self._expand()
        return self.final_bpe_seq


class SamplingBytePairEncoder:
    """
    Will randomly sample from any possible BPE split.
    """

    def __init__(self, labels, breadth_prob, rnd, unknown_label=None):
        """
        :param list[str] labels: vocab
        :param float breadth_prob: 1.0 will lead to breadth-first search, 0.0 to depth-first search.
          other values are stochastic.
        :param numpy.random.RandomState rnd:
        :param str|None unknown_label:
        """
        self.labels = labels
        self.unknown_label = unknown_label
        if unknown_label:
            assert unknown_label in self.labels
        self.breadth_prob = breadth_prob
        self.rnd = rnd

        # build prefix tree
        bpe = PrefixTree()
        for bpe_sym in labels:
            bpe.add(bpe_sym)
        self._bpe_prefix_tree = bpe

    def _sampler(self) -> bool:
        # When this returns true, it will differ from depth-first search.
        return self.rnd.random_sample() <= self.breadth_prob

    def get_bpe_split_for_word(self, word: str) -> Optional[List[str]]:
        """
        :param word:
        """
        return DepthFirstSearch(bpe=self._bpe_prefix_tree, word=word, sampler=self._sampler).search()

    def segment_sentence(self, sentence: str) -> List[str]:
        """
        Segment single sentence (whitespace-tokenized string) with BPE encoding.

        :param sentence:
        """
        output = []
        for word in sentence.split():
            bpe_sym_seq = self.get_bpe_split_for_word(word)
            if bpe_sym_seq is None:
                if self.unknown_label:
                    output.append(self.unknown_label)
                    continue
                else:
                    raise Exception("no BPE-split for word %r" % word)
            output.extend(bpe_sym_seq)
        return output


def _demo():
    import sys
    import os

    my_dir = os.path.dirname(os.path.abspath(__file__))
    root_dir = os.path.dirname(os.path.dirname(my_dir))
    assert os.path.exists("%s/returnn/__init__.py" % root_dir)
    sys.path.insert(0, root_dir)

    from returnn.util import better_exchook

    better_exchook.install()

    import argparse

    arg_parser = argparse.ArgumentParser()
    arg_parser.add_argument("--vocab", required=True)
    arg_parser.add_argument("--unk")
    arg_parser.add_argument("--input", help="text. if not given, will read from stdin")
    arg_parser.add_argument("--seed", type=int, default=0)
    arg_parser.add_argument("--all", action="store_true")
    arg_parser.add_argument(
        "--breadth-prob",
        type=float,
        default=0.0,
        help="1.0 will lead to breadth-first search, 0.0 to depth-first search. other values are stochastic.",
    )
    args = arg_parser.parse_args()

    from returnn.datasets.util.vocabulary import Vocabulary

    vocab = Vocabulary(vocab_file=args.vocab, unknown_label=None)
    rnd = numpy.random.RandomState(args.seed)

    if args.input:
        bpe_prefix_tree = PrefixTree()
        for bpe_sym in vocab.labels:
            bpe_prefix_tree.add(bpe_sym)

        def _sampler():
            # When this returns true, it will differ from depth-first search.
            return rnd.random_sample() <= args.breadth_prob

        for word in args.input.split():
            if args.all:
                bpe_sym_seqs = CharSyncSearch(bpe=bpe_prefix_tree, word=word).search()
                print("%s: %s" % (word, bpe_sym_seqs))
            else:
                greedy = DepthFirstSearch(bpe=bpe_prefix_tree, word=word, sampler=_sampler).search()
                print("%s: %s" % (word, " ".join(greedy)))
        return

    bpe = SamplingBytePairEncoder(labels=vocab.labels, breadth_prob=args.breadth_prob, rnd=rnd, unknown_label=args.unk)
    print("Reading from stdin:")
    while True:
        try:
            line = sys.stdin.readline()
            if line == "":  # EOF
                return
        except KeyboardInterrupt:
            return
        line = line.strip()
        print(" ".join(bpe.segment_sentence(line)))


if __name__ == "__main__":
    _demo()



"""
Construct modules (or other objects) from dictionaries.
"""

from __future__ import annotations
from typing import Union, Any, Type, Dict
import importlib
import returnn.frontend as rf


__all__ = ["build_from_dict", "build_dict"]


def build_from_dict(d: Dict[str, Any], *args, **kwargs) -> Union[rf.Module, Any]:
    """
    Build a module (or other object) from a dictionary.
    `"class"` in the dict is required and specifies the class to be instantiated.
    The other options are passed to the class constructor.

    :param d: dictionary with the class name and other options
    :param args: passed to the class constructor
    :param kwargs: passed to the class constructor
    :return: cls(*args, **d, **kwargs)
    """
    if "class" not in d:
        raise ValueError(f"build_from_dict: Missing 'class' key in dict: {d}")
    d = d.copy()
    cls_name = d.pop("class")
    cls = _get_cls(cls_name)
    return cls(*args, **d, **kwargs)


def build_dict(cls: Type, **kwargs) -> Dict[str, Any]:
    """
    Build a dictionary for :func:`build_from_dict`.
    The class name is stored in the `"class"` key.

    Note that this is intended to be used for serialization
    and also to get a unique stable hashable representation
    (e.g. for Sisyphus :func:`sis_hash_helper`)
    which should not change if the class is renamed or moved
    to keep the hash stable.
    """
    return {"class": _get_cls_name(cls), **kwargs}


def _get_cls(cls_name: str) -> Type:
    if "." not in cls_name:
        raise ValueError(f"Expected '.' in class name: {cls_name}")
    mod_name, cls_name = cls_name.rsplit(".", 1)
    if mod_name == "rf":
        return getattr(rf, cls_name)
    mod = importlib.import_module(mod_name)
    return getattr(mod, cls_name)


def _get_cls_name(cls: Type) -> str:
    if getattr(rf, cls.__name__, None) is cls:
        return f"rf.{cls.__name__}"
    return f"{cls.__module__}.{cls.__name__}"



# Copyright 2015 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Various ways of selecting operations and tensors in a graph."""

from __future__ import annotations

import re

from six import iteritems
from six import string_types

from . import util
import tensorflow as tf
from tensorflow.python.framework import ops as tf_ops


__all__ = [
    "can_be_regex",
    "make_regex",
    "filter_ts",
    "filter_ts_from_regex",
    "filter_ops",
    "filter_ops_from_regex",
    "get_name_scope_ops",
    "check_cios",
    "get_ops_ios",
    "compute_boundary_ts",
    "get_within_boundary_ops",
    "get_forward_walk_ops",
    "get_backward_walk_ops",
    "get_walks_intersection_ops",
    "get_walks_union_ops",
    "select_ops",
    "select_ts",
    "select_ops_and_ts",
]

_RE_TYPE = type(re.compile(""))


def can_be_regex(obj):
    """Return True if obj can be turned into a regular expression."""
    return isinstance(obj, string_types + (_RE_TYPE,))


def make_regex(obj):
    """Return a compiled regular expression.

    Args:
      obj: a string or a regular expression.
    Returns:
      A compiled regular expression.
    Raises:
      ValueError: if obj could not be converted to a regular expression.
    """
    if not can_be_regex(obj):
        raise ValueError("Expected a string or a regex, got: {}".format(type(obj)))

    if isinstance(obj, string_types):
        return re.compile(obj)
    else:
        return obj


def _get_input_ts(ops):
    """Compute the list of unique input tensors of all the op in ops.

    Args:
      ops: an object convertible to a list of `tf.Operation`.
    Returns:
      The list of unique input tensors of all the op in ops.
    Raises:
      TypeError: if ops cannot be converted to a list of `tf.Operation`.
    """
    ops = util.make_list_of_op(ops)
    ts = []
    ts_set = set()
    for op in ops:
        for t in op.inputs:
            if t not in ts_set:
                ts.append(t)
                ts_set.add(t)
    return ts


def _get_output_ts(ops):
    """Compute the list of unique output tensors of all the op in ops.

    Args:
      ops: an object convertible to a list of tf.Operation.
    Returns:
      The list of unique output tensors of all the op in ops.
    Raises:
      TypeError: if ops cannot be converted to a list of tf.Operation.
    """
    ops = util.make_list_of_op(ops)
    ts = []
    for op in ops:
        ts += op.outputs
    return ts


def filter_ts(ops, positive_filter):
    """Get all the tensors which are input or output of an op in ops.

    Args:
      ops: an object convertible to a list of `tf.Operation`.
      positive_filter: a function deciding whether to keep a tensor or not.
        If `True`, all the tensors are returned.
    Returns:
      A list of `tf.Tensor`.
    Raises:
      TypeError: if ops cannot be converted to a list of `tf.Operation`.
    """
    ops = util.make_list_of_op(ops)
    ts = _get_input_ts(ops)
    util.concatenate_unique(ts, _get_output_ts(ops))
    if positive_filter is not True:
        ts = [t for t in ts if positive_filter(t)]
    return ts


def filter_ts_from_regex(ops, regex):
    r"""Get all the tensors linked to ops that match the given regex.

    Args:
      ops: an object convertible to a list of tf.Operation.
      regex: a regular expression matching the tensors' name.
        For example, "^foo(/.*)?:\d+$" will match all the tensors in the "foo"
        scope.
    Returns:
      A list of tf.Tensor.
    Raises:
      TypeError: if ops cannot be converted to a list of tf.Operation.
    """
    ops = util.make_list_of_op(ops)
    regex_obj = make_regex(regex)
    return filter_ts(ops, positive_filter=lambda op: regex_obj.search(op.name))


def filter_ops(ops, positive_filter):
    """Get the ops passing the given filter.

    Args:
      ops: an object convertible to a list of tf.Operation.
      positive_filter: a function deciding where to keep an operation or not.
        If True, all the operations are returned.
    Returns:
      A list of selected tf.Operation.
    Raises:
      TypeError: if ops cannot be converted to a list of tf.Operation.
    """
    ops = util.make_list_of_op(ops)
    if positive_filter is not True:  # pylint: disable=g-explicit-bool-comparison
        ops = [op for op in ops if positive_filter(op)]
    return ops


def filter_ops_from_regex(ops, regex):
    """Get all the operations that match the given regex.

    Args:
      ops: an object convertible to a list of `tf.Operation`.
      regex: a regular expression matching the operation's name.
        For example, `"^foo(/.*)?$"` will match all the operations in the "foo"
        scope.
    Returns:
      A list of `tf.Operation`.
    Raises:
      TypeError: if ops cannot be converted to a list of `tf.Operation`.
    """
    ops = util.make_list_of_op(ops)
    regex_obj = make_regex(regex)
    return filter_ops(ops, lambda op: regex_obj.search(op.name))


def get_name_scope_ops(ops, scope):
    """Get all the operations under the given scope path.

    Args:
      ops: an object convertible to a list of tf.Operation.
      scope: a scope path.
    Returns:
      A list of tf.Operation.
    Raises:
      TypeError: if ops cannot be converted to a list of tf.Operation.
    """
    if scope and scope[-1] == "/":
        scope = scope[:-1]
    return filter_ops_from_regex(ops, "^{}(/.*)?$".format(scope))


def check_cios(control_inputs=False, control_outputs=None, control_ios=None):
    """Do various check on control_inputs and control_outputs.

    Args:
      control_inputs: A boolean indicating whether control inputs are enabled.
      control_outputs: An instance of util.ControlOutputs or None. If not None,
        control outputs are enabled.
      control_ios:  An instance of util.ControlOutputs or None. If not None, both
        control inputs and control outputs are enabled. This is equivalent to set
        control_inputs to True and control_outputs to the util.ControlOutputs
        instance.
    Returns:
      A tuple `(control_inputs, control_outputs)` where:
        `control_inputs` is a boolean indicating whether to use control inputs.
        `control_outputs` is an instance of util.ControlOutputs or None
    Raises:
      ValueError: if control_inputs is an instance of util.ControlOutputs but
        control_outputs is not None
      TypeError: if control_outputs is not None and is not a util.ControlOutputs.
    """
    if control_ios is not None:
        if not isinstance(control_ios, util.ControlOutputs):
            raise TypeError("Expected a util.ControlOutputs, got: {}".format(type(control_ios)))
        if control_outputs is not None:
            raise ValueError("control_outputs should be None when using control_ios.")
        control_inputs = True
        control_outputs = control_ios
    elif control_outputs is not None:
        if not isinstance(control_outputs, util.ControlOutputs):
            raise TypeError("Expected a util.ControlOutputs, got: {}".format(type(control_outputs)))

    if control_outputs is not None:
        control_outputs.update()
    return control_inputs, control_outputs


def get_ops_ios(ops, control_inputs=False, control_outputs=None, control_ios=None):
    """Return all the `tf.Operation` which are connected to an op in ops.

    Args:
      ops: an object convertible to a list of `tf.Operation`.
      control_inputs: A boolean indicating whether control inputs are enabled.
      control_outputs: An instance of `util.ControlOutputs` or `None`. If not
        `None`, control outputs are enabled.
      control_ios:  An instance of `util.ControlOutputs` or `None`. If not `None`,
        both control inputs and control outputs are enabled. This is equivalent to
        set `control_inputs` to `True` and `control_outputs` to the
        `util.ControlOutputs` instance.
    Returns:
      All the `tf.Operation` surrounding the given ops.
    Raises:
      TypeError: if `ops` cannot be converted to a list of `tf.Operation`.
    """
    control_inputs, control_outputs = check_cios(control_inputs, control_outputs, control_ios)
    ops = util.make_list_of_op(ops)
    res = []
    for op in ops:
        util.concatenate_unique(res, [t.op for t in op.inputs])
        for t in op.outputs:
            util.concatenate_unique(res, t.consumers())
        if control_outputs is not None:
            util.concatenate_unique(res, control_outputs.get(op))
        if control_inputs:
            util.concatenate_unique(res, op.control_inputs)
    return res


def compute_boundary_ts(ops):
    """Compute the tensors at the boundary of a set of ops.

    This function looks at all the tensors connected to the given ops (in/out)
    and classify them into three categories:
    1) input tensors: tensors whose generating operation is not in ops.
    2) output tensors: tensors whose consumer operations are not in ops
    3) inside tensors: tensors which are neither input nor output tensors.

    Note that a tensor can be both an inside tensor and an output tensor if it is
    consumed by operations both outside and inside of `ops`.

    Args:
      ops: an object convertible to a list of tf.Operation.
    Returns:
      A tuple `(outside_input_ts, outside_output_ts, inside_ts)` where:
        `outside_input_ts` is a Python list of input tensors;
        `outside_output_ts` is a python list of output tensors;
        `inside_ts` is a python list of inside tensors.
      Since a tensor can be both an inside tensor and an output tensor,
      `outside_output_ts` and `inside_ts` might intersect.
    Raises:
      TypeError: if ops cannot be converted to a list of tf.Operation.
    """
    ops = util.make_list_of_op(ops)
    input_ts = _get_input_ts(ops)
    output_ts = _get_output_ts(ops)
    output_ts_set = frozenset(output_ts)
    ops_set = frozenset(ops)

    # Compute inside tensors.
    inside_ts = []
    only_inside_ts = []
    for t in input_ts:
        # Skip if the input tensor is not also an output tensor.
        if t not in output_ts_set:
            continue
        # Mark as "inside".
        inside_ts.append(t)
        # Mark as "only inside" if the tensor is not both inside and output.
        consumers = frozenset(t.consumers())
        if consumers - ops_set:
            continue
        only_inside_ts.append(t)

    inside_ts_set = frozenset(inside_ts)
    only_inside_ts_set = frozenset(only_inside_ts)
    outside_output_ts = [t for t in output_ts if t not in only_inside_ts_set]
    outside_input_ts = [t for t in input_ts if t not in inside_ts_set]
    return outside_input_ts, outside_output_ts, inside_ts


def get_within_boundary_ops(
    ops, seed_ops, boundary_ops=(), inclusive=True, control_inputs=False, control_outputs=None, control_ios=None
):
    """Return all the `tf.Operation` within the given boundary.

    Args:
      ops: an object convertible to a list of `tf.Operation`. those ops define the
        set in which to perform the operation (if a `tf.Graph` is given, it
        will be converted to the list of all its operations).
      seed_ops: the operations from which to start expanding.
      boundary_ops: the ops forming the boundary.
      inclusive: if `True`, the result will also include the boundary ops.
      control_inputs: A boolean indicating whether control inputs are enabled.
      control_outputs: An instance of `util.ControlOutputs` or `None`. If not
        `None`, control outputs are enabled.
      control_ios:  An instance of `util.ControlOutputs` or `None`. If not
        `None`, both control inputs and control outputs are enabled. This is
        equivalent to set control_inputs to True and control_outputs to
        the `util.ControlOutputs` instance.
    Returns:
      All the `tf.Operation` surrounding the given ops.
    Raises:
      TypeError: if `ops` or `seed_ops` cannot be converted to a list of
        `tf.Operation`.
      ValueError: if the boundary is intersecting with the seeds.
    """
    control_inputs, control_outputs = check_cios(control_inputs, control_outputs, control_ios)
    ops = util.make_list_of_op(ops)
    seed_ops = util.make_list_of_op(seed_ops, allow_graph=False)
    boundary_ops = set(util.make_list_of_op(boundary_ops))
    res = set(seed_ops)
    if boundary_ops & res:
        raise ValueError("Boundary is intersecting with the seeds.")
    wave = set(seed_ops)
    while wave:
        new_wave = set()
        ops_io = get_ops_ios(wave, control_inputs, control_outputs)
        for op in ops_io:
            if op in res:
                continue
            if op in boundary_ops:
                if inclusive:
                    res.add(op)
            else:
                new_wave.add(op)
        res.update(new_wave)
        wave = new_wave
    return [op for op in ops if op in res]


def get_forward_walk_ops(
    seed_ops, inclusive=True, within_ops=None, within_ops_fn=None, stop_at_ts=(), control_outputs=None
):
    """Do a forward graph walk and return all the visited ops.

    Args:
      seed_ops: an iterable of operations from which the forward graph
        walk starts. If a list of tensors is given instead, the seed_ops are set
        to be the consumers of those tensors.
      inclusive: if True the given seed_ops are also part of the resulting set.
      within_ops: an iterable of `tf.Operation` within which the search is
        restricted. If `within_ops` is `None`, the search is performed within
        the whole graph.
      within_ops_fn: if provided, a function on ops that should return True iff
        the op is within the graph traversal. This can be used along within_ops,
        in which case an op is within if it is also in within_ops.
      stop_at_ts: an iterable of tensors at which the graph walk stops.
      control_outputs: a `util.ControlOutputs` instance or None.
        If not `None`, it will be used while walking the graph forward.
    Returns:
      A Python set of all the `tf.Operation` ahead of `seed_ops`.
    Raises:
      TypeError: if `seed_ops` or `within_ops` cannot be converted to a list of
        `tf.Operation`.
    """
    _, control_outputs = check_cios(False, control_outputs)
    if not util.is_iterable(seed_ops):
        seed_ops = [seed_ops]
    if not seed_ops:
        return []
    if isinstance(seed_ops[0], tf.Tensor):
        ts = util.make_list_of_t(seed_ops, allow_graph=False)
        seed_ops = util.get_consuming_ops(ts)
    else:
        seed_ops = util.make_list_of_op(seed_ops, allow_graph=False)

    seed_ops = frozenset(seed_ops)
    stop_at_ts = frozenset(util.make_list_of_t(stop_at_ts))
    if within_ops:
        within_ops = util.make_list_of_op(within_ops, allow_graph=False)
        within_ops = frozenset(within_ops)
        seed_ops &= within_ops

    # noinspection PyShadowingNames
    def is_within(op):
        """
        :param tf.Operation op:
        :rtype: bool
        """
        return (within_ops is None or op in within_ops) and (within_ops_fn is None or within_ops_fn(op))

    result = list(seed_ops)
    wave = set(seed_ops)
    while wave:
        new_wave = set()
        for op in wave:
            for new_t in op.outputs:
                if new_t in stop_at_ts:
                    continue
                for new_op in new_t.consumers():
                    if new_op not in result and is_within(new_op):
                        new_wave.add(new_op)
            if control_outputs is not None:
                for new_op in control_outputs.get(op):
                    if new_op not in result and is_within(new_op):
                        new_wave.add(new_op)
        util.concatenate_unique(result, new_wave)
        wave = new_wave
    if not inclusive:
        result = [op for op in result if op not in seed_ops]
    return result


def get_backward_walk_ops(
    seed_ops, inclusive=True, within_ops=None, within_ops_fn=None, stop_at_ts=(), control_inputs=False
):
    """Do a backward graph walk and return all the visited ops.

    Args:
      seed_ops: an iterable of operations from which the backward graph
        walk starts. If a list of tensors is given instead, the seed_ops are set
        to be the generators of those tensors.
      inclusive: if True the given seed_ops are also part of the resulting set.
      within_ops: an iterable of `tf.Operation` within which the search is
        restricted. If `within_ops` is `None`, the search is performed within
        the whole graph.
      within_ops_fn: if provided, a function on ops that should return True iff
        the op is within the graph traversal. This can be used along within_ops,
        in which case an op is within if it is also in within_ops.
      stop_at_ts: an iterable of tensors at which the graph walk stops.
      control_inputs: if True, control inputs will be used while moving backward.
    Returns:
      A Python set of all the `tf.Operation` behind `seed_ops`.
    Raises:
      TypeError: if `seed_ops` or `within_ops` cannot be converted to a list of
        `tf.Operation`.
    """
    try:
        from tensorflow.contrib.graph_editor import select as op_selector
    except ImportError:
        from tensorflow.python.ops import op_selector
    return op_selector.get_backward_walk_ops(
        seed_ops,
        inclusive=inclusive,
        within_ops=within_ops,
        within_ops_fn=within_ops_fn,
        stop_at_ts=stop_at_ts,
        control_inputs=control_inputs,
    )


def get_walks_intersection_ops(
    forward_seed_ops,
    backward_seed_ops,
    forward_inclusive=True,
    backward_inclusive=True,
    within_ops=None,
    within_ops_fn=None,
    control_inputs=False,
    control_outputs=None,
    control_ios=None,
):
    """Return the intersection of a forward and a backward walk.

    Args:
      forward_seed_ops: an iterable of operations from which the forward graph
        walk starts. If a list of tensors is given instead, the seed_ops are set
        to be the consumers of those tensors.
      backward_seed_ops: an iterable of operations from which the backward graph
        walk starts. If a list of tensors is given instead, the seed_ops are set
        to be the generators of those tensors.
      forward_inclusive: if True the given forward_seed_ops are also part of the
        resulting set.
      backward_inclusive: if True the given backward_seed_ops are also part of the
        resulting set.
      within_ops: an iterable of tf.Operation within which the search is
        restricted. If within_ops is None, the search is performed within
        the whole graph.
      within_ops_fn: if provided, a function on ops that should return True iff
        the op is within the graph traversal. This can be used along within_ops,
        in which case an op is within if it is also in within_ops.
      control_inputs: A boolean indicating whether control inputs are enabled.
      control_outputs: An instance of util.ControlOutputs or None. If not None,
        control outputs are enabled.
      control_ios:  An instance of util.ControlOutputs or None. If not None, both
        control inputs and control outputs are enabled. This is equivalent to set
        control_inputs to True and control_outputs to the util.ControlOutputs
        instance.
    Returns:
      A Python set of all the tf.Operation in the intersection of a forward and a
        backward walk.
    Raises:
      TypeError: if `forward_seed_ops` or `backward_seed_ops` or `within_ops`
        cannot be converted to a list of `tf.Operation`.
    """
    control_inputs, control_outputs = check_cios(control_inputs, control_outputs, control_ios)
    forward_ops = get_forward_walk_ops(
        forward_seed_ops,
        inclusive=forward_inclusive,
        within_ops=within_ops,
        within_ops_fn=within_ops_fn,
        control_outputs=control_outputs,
    )
    backward_ops = get_backward_walk_ops(
        backward_seed_ops,
        inclusive=backward_inclusive,
        within_ops=within_ops,
        within_ops_fn=within_ops_fn,
        control_inputs=control_inputs,
    )
    return [op for op in forward_ops if op in backward_ops]


def get_walks_union_ops(
    forward_seed_ops,
    backward_seed_ops,
    forward_inclusive=True,
    backward_inclusive=True,
    within_ops=None,
    within_ops_fn=None,
    control_inputs=False,
    control_outputs=None,
    control_ios=None,
):
    """Return the union of a forward and a backward walk.

    Args:
      forward_seed_ops: an iterable of operations from which the forward graph
        walk starts. If a list of tensors is given instead, the seed_ops are set
        to be the consumers of those tensors.
      backward_seed_ops: an iterable of operations from which the backward graph
        walk starts. If a list of tensors is given instead, the seed_ops are set
        to be the generators of those tensors.
      forward_inclusive: if True the given forward_seed_ops are also part of the
        resulting set.
      backward_inclusive: if True the given backward_seed_ops are also part of the
        resulting set.
      within_ops: restrict the search within those operations. If within_ops is
        None, the search is done within the whole graph.
      within_ops_fn: if provided, a function on ops that should return True iff
        the op is within the graph traversal. This can be used along within_ops,
        in which case an op is within if it is also in within_ops.
      control_inputs: A boolean indicating whether control inputs are enabled.
      control_outputs: An instance of util.ControlOutputs or None. If not None,
        control outputs are enabled.
      control_ios:  An instance of util.ControlOutputs or None. If not None, both
        control inputs and control outputs are enabled. This is equivalent to set
        control_inputs to True and control_outputs to the util.ControlOutputs
        instance.
    Returns:
      A Python set of all the tf.Operation in the union of a forward and a
        backward walk.
    Raises:
      TypeError: if forward_seed_ops or backward_seed_ops or within_ops cannot be
        converted to a list of tf.Operation.
    """
    control_inputs, control_outputs = check_cios(control_inputs, control_outputs, control_ios)
    forward_ops = get_forward_walk_ops(
        forward_seed_ops,
        inclusive=forward_inclusive,
        within_ops=within_ops,
        within_ops_fn=within_ops_fn,
        control_outputs=control_outputs,
    )
    backward_ops = get_backward_walk_ops(
        backward_seed_ops,
        inclusive=backward_inclusive,
        within_ops=within_ops,
        within_ops_fn=within_ops_fn,
        control_inputs=control_inputs,
    )
    return util.concatenate_unique(forward_ops, backward_ops)


def select_ops(*args, **kwargs):
    """Helper to select operations.

    Args:
      *args: list of 1) regular expressions (compiled or not) or 2) (array of)
        `tf.Operation`. `tf.Tensor` instances are silently ignored.
      **kwargs: 'graph': `tf.Graph` in which to perform the regex query.This is
        required when using regex.
        'positive_filter': an elem if selected only if `positive_filter(elem)` is
          `True`. This is optional.
        'restrict_ops_regex': a regular expression is ignored if it doesn't start
          with the substring "(?#ops)".
    Returns:
      A list of `tf.Operation`.
    Raises:
      TypeError: if the optional keyword argument graph is not a `tf.Graph`
        or if an argument in args is not an (array of) `tf.Operation`
        or an (array of) `tf.Tensor` (silently ignored) or a string
        or a regular expression.
      ValueError: if one of the keyword arguments is unexpected or if a regular
        expression is used without passing a graph as a keyword argument.
    """
    # get keywords arguments
    graph = None
    positive_filter = None
    restrict_ops_regex = False
    for k, v in iteritems(kwargs):
        if k == "graph":
            graph = v
            if graph is not None and not isinstance(graph, tf_ops.Graph):
                raise TypeError("Expected a tf.Graph, got: {}".format(type(graph)))
        elif k == "positive_filter":
            positive_filter = v
        elif k == "restrict_ops_regex":
            restrict_ops_regex = v
        elif k == "restrict_ts_regex":
            pass
        else:
            raise ValueError("Wrong keywords argument: {}.".format(k))

    ops = []

    for arg in args:
        if can_be_regex(arg):
            if graph is None:
                raise ValueError("Use the keyword argument 'graph' to use regex.")
            regex = make_regex(arg)
            if regex.pattern.startswith("(?#ts)"):
                continue
            if restrict_ops_regex and not regex.pattern.startswith("(?#ops)"):
                continue
            ops_ = filter_ops_from_regex(graph, regex)
            for op_ in ops_:
                if op_ not in ops:
                    if positive_filter is None or positive_filter(op_):
                        ops.append(op_)
        else:
            ops_aux = util.make_list_of_op(arg, ignore_ts=True)
            if positive_filter is not None:
                ops_aux = [op for op in ops_aux if positive_filter(op)]
            ops_aux = [op for op in ops_aux if op not in ops]
            ops += ops_aux

    return ops


def select_ts(*args, **kwargs):
    """Helper to select tensors.

    Args:
      *args: list of 1) regular expressions (compiled or not) or 2) (array of)
        `tf.Tensor`. `tf.Operation` instances are silently ignored.
      **kwargs: 'graph': `tf.Graph` in which to perform the regex query.This is
        required when using regex.
        'positive_filter': an elem if selected only if `positive_filter(elem)` is
          `True`. This is optional.
        'restrict_ts_regex': a regular expression is ignored if it doesn't start
          with the substring "(?#ts)".
    Returns:
      A list of `tf.Tensor`.
    Raises:
      TypeError: if the optional keyword argument graph is not a `tf.Graph`
        or if an argument in args is not an (array of) `tf.Tensor`
        or an (array of) `tf.Operation` (silently ignored) or a string
        or a regular expression.
      ValueError: if one of the keyword arguments is unexpected or if a regular
        expression is used without passing a graph as a keyword argument.
    """
    # get keywords arguments
    graph = None
    positive_filter = None
    restrict_ts_regex = False
    for k, v in iteritems(kwargs):
        if k == "graph":
            graph = v
            if graph is not None and not isinstance(graph, tf_ops.Graph):
                raise TypeError("Expected a tf.Graph, got {}".format(type(graph)))
        elif k == "positive_filter":
            positive_filter = v
        elif k == "restrict_ts_regex":
            restrict_ts_regex = v
        elif k == "restrict_ops_regex":
            pass
        else:
            raise ValueError("Wrong keywords argument: {}.".format(k))

    ts = []

    for arg in args:
        if can_be_regex(arg):
            if graph is None:
                raise ValueError("Use the keyword argument 'graph' to use regex.")
            regex = make_regex(arg)
            if regex.pattern.startswith("(?#ops)"):
                continue
            if restrict_ts_regex and not regex.pattern.startswith("(?#ts)"):
                continue
            ts_ = filter_ts_from_regex(graph, regex)
            for t_ in ts_:
                if t_ not in ts:
                    if positive_filter is None or positive_filter(t_):
                        ts.append(t_)
        else:
            ts_aux = util.make_list_of_t(arg, ignore_ops=True)
            if positive_filter is not None:
                ts_aux = [t for t in ts_aux if positive_filter(t)]
            ts_aux = [t for t in ts_aux if t not in ts]
            ts += ts_aux

    return ts


def select_ops_and_ts(*args, **kwargs):
    """Helper to select operations and tensors.

    Args:
      *args: list of 1) regular expressions (compiled or not) or 2) (array of)
        `tf.Operation` 3) (array of) tf.Tensor. Regular expressions matching
        tensors must start with the comment `"(?#ts)"`, for instance:
        `"(?#ts)^foo/.*"`.
      **kwargs: 'graph': `tf.Graph` in which to perform the regex query.This is
        required when using regex.
        'positive_filter': an elem if selected only if `positive_filter(elem)` is
          `True`. This is optional.
    Returns:
      A tuple `(ops, ts)` where:
        `ops` is a list of `tf.Operation`, and
        `ts` is a list of `tf.Tensor`
    Raises:
      TypeError: if the optional keyword argument graph is not a `tf.Graph`
        or if an argument in args is not an (array of) `tf.Tensor`
        or an (array of) `tf.Operation` or a string or a regular expression.
      ValueError: if one of the keyword arguments is unexpected or if a regular
        expression is used without passing a graph as a keyword argument.
    """
    ops = select_ops(*args, restrict_ops_regex=False, **kwargs)
    ts = select_ts(*args, restrict_ts_regex=True, **kwargs)
    return ops, ts



#!/usr/bin/env python

"""
Iterate through a dataset, just like RETURNN would do in training.
"""

from __future__ import annotations

import sys
import typing

import _setup_returnn_env  # noqa
from returnn import __main__ as rnn
from returnn.log import log
from returnn.engine.base import EngineBase

dev_num_batches = 1


def iterate_dataset(dataset, recurrent_net, batch_size, max_seqs):
    """
    :type dataset: Dataset.Dataset
    :type recurrent_net: bool
    :type batch_size: int
    :type max_seqs: int
    """
    batch_gen = dataset.generate_batches(recurrent_net=recurrent_net, batch_size=batch_size, max_seqs=max_seqs)
    while batch_gen.has_more():
        batches = batch_gen.peek_next_n(dev_num_batches)
        for batch in batches:
            dataset.load_seqs(batch.start_seq, batch.end_seq)
        batch_gen.advance(len(batches))


def iterate_epochs():
    """
    Iterate through epochs.
    """
    start_epoch = 1
    final_epoch = EngineBase.config_get_final_epoch(config)

    print("Starting with epoch %i." % (start_epoch,), file=log.v3)
    print("Final epoch is: %i" % final_epoch, file=log.v3)

    recurrent_net = "lstm" in config.value("hidden_type", "")  # good enough...
    batch_size = config.int("batch_size", 1)
    max_seqs = config.int("max_seqs", -1)

    for epoch in range(start_epoch, final_epoch + 1):
        print("Epoch %i." % epoch, file=log.v3)
        rnn.train_data.init_seq_order(epoch)
        iterate_dataset(rnn.train_data, recurrent_net=recurrent_net, batch_size=batch_size, max_seqs=max_seqs)

    print("Finished all epochs.", file=log.v3)


config = None  # type: typing.Optional["returnn.config.Config"]


def init(config_filename, command_line_options):
    """
    :param str config_filename:
    :param list[str] command_line_options:
    """
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    rnn.init_config(config_filename, command_line_options)
    global config
    config = rnn.config
    rnn.init_log()
    print("RETURNN demo-dataset starting up", file=log.v3)
    rnn.init_faulthandler()
    rnn.init_data()
    rnn.print_task_properties()


def main(argv):
    """
    Main entry.
    """
    assert len(argv) >= 2, "usage: %s <config>" % argv[0]
    init(config_filename=argv[1], command_line_options=argv[2:])
    iterate_epochs()
    rnn.finalize()


if __name__ == "__main__":
    main(sys.argv)
